#
#ifndef MCPGEN_H
#define MCPGEN_H

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <inttypes.h>
#include <getopt.h>
#include <time.h>

#define IP_HEADER_SIZE 20

struct options
{
	char mcast_addr[17];
	int port;
	int pkt_size;
	int pkts_per_sec;
	int ttl;
	int burst;
	int buffer_size;
	int max_clients;
};

struct message
{
	uint64_t seq_number;
	uint64_t pkts_per_sec;
};

struct client_data
{
	int naddr;
	char addr[17];
	int port;
	uint64_t last_seq;
	uint64_t first_seq;
	uint64_t recieved_pkts;
	uint64_t pkts_per_sec;
	uint64_t late_pkts;
	uint64_t early_pkts;
};

extern int usage( char *prog );

int get_options( int argc, char * const arrgv[], struct options *opts );

int open_listener_socket( char *addr, int port, int buffer_size );

int open_sender_socket( int port, int ttl, int buffer_size );
int setup_dest_addr( struct sockaddr_in *sockaddr, char *addr, int port );

struct client_data *find_client( int naddr, struct client_data *clients, int max );

int open_listener_socket( char *addr, int port, int buffer_size )
{
	int sd;
	int len;
	int tmp;
	struct hostent *hostent;
	struct in_addr in_addr;
	struct sockaddr_in sockaddr;
	struct ip_mreq mreq;

	/* Attempt name resolution */
	hostent = gethostbyname(addr);
	if ( hostent == NULL )
	{
		printf ( "could not resolve '%s'\n", addr );
		return -1;
	}

	/* use first returned address */
	memcpy( &in_addr, hostent->h_addr_list[0], hostent->h_length );
	
	/* check that the addr is in mcast range */
	if ( ! IN_MULTICAST(ntohl(in_addr.s_addr)) )
	{
		printf ( "'%s' not a valid mcast group addr\n", addr );
		return -2;
	}

	/* create our socket */
	sd = socket(AF_INET,SOCK_DGRAM,0);
	if ( sd<0 )
	{
		printf ( "failed to create socket\n" );
		return -3;
	}

	/* bind to port */
	sockaddr.sin_family=AF_INET;
	sockaddr.sin_addr.s_addr=htonl(INADDR_ANY);
	sockaddr.sin_port=htons(port);
	if( bind( sd, (struct sockaddr *) &sockaddr, sizeof(sockaddr)) < 0 )
	{
		printf ( "failed to bind to port %d\n", port );
		return -4;
	}

	/* bind to mcast addr */
	mreq.imr_multiaddr.s_addr=in_addr.s_addr;
	mreq.imr_interface.s_addr=htonl(INADDR_ANY);

	if ( setsockopt( sd, IPPROTO_IP,IP_ADD_MEMBERSHIP, (void *) &mreq, sizeof(mreq) ) < 0 )
	{
		printf ( "failed to join group '%s'\n", addr );
		return -5;
	}

	/* set the recv buffer size */
	if ( buffer_size != 0 )
	{
		if( setsockopt( sd, SOL_SOCKET, SO_RCVBUF, &buffer_size, sizeof(buffer_size) ) < 0  )
		{
			printf ( "failed to set rcv buffer to %d\n", buffer_size  );
			return -6;
		}
	}


	len = sizeof(tmp);
	getsockopt( sd, SOL_SOCKET, SO_RCVBUF, &tmp, &len );

	printf ("listening to %s on port %d (rcv buffer size %d bytes)\n", inet_ntoa(in_addr), port, tmp ); 
	return sd;
}



int open_sender_socket( int port, int ttl, int buffer_size )
{
	int sd, len;
	int tmp;
	struct sockaddr_in sockaddr;

	/* create socket */
	sd = socket(AF_INET,SOCK_DGRAM,0);
	if ( sd<0 )
	{
		printf ( "failed to create socket\n" );
		return -3;
	}

	/* bind to port */
	sockaddr.sin_family=AF_INET;
	sockaddr.sin_addr.s_addr=htonl(INADDR_ANY);
	sockaddr.sin_port=htons(port);
	if( bind( sd, (struct sockaddr *) &sockaddr, sizeof(sockaddr)) < 0 )
	{
		printf ( "failed to bind to port %d\n", port );
		return -4;
	}

	/* set the mcast ttl on the packets */
	if ( ttl > 0 ) /* if passed 0, just accept the system dflt */
	{
		if( setsockopt(sd,IPPROTO_IP,IP_MULTICAST_TTL, &ttl,sizeof(ttl) ) < 0 )
		{
			printf ( "failed to set ttl to %d\n", ttl );
			return -5;
		}
	}

	

	len = sizeof(sockaddr);
	getsockname( sd, (struct sockaddr *) &sockaddr, &len );
	port = ntohs(sockaddr.sin_port);

	/* set the send buffer size */
	if ( buffer_size != 0 )
	{
		if( setsockopt( sd, SOL_SOCKET, SO_SNDBUF, &buffer_size, sizeof(buffer_size) ) < 0  )
		{
			printf ( "failed to set send buffer to %d\n", buffer_size  );
			return -6;
		}
	}


	len = sizeof(tmp);
	getsockopt( sd, SOL_SOCKET, SO_SNDBUF, &tmp, &len );

	printf( "bound to %s port %d (send buffer size %d bytes)\n", inet_ntoa(sockaddr.sin_addr), port, tmp );
	return sd;
}

int setup_dest_addr( struct sockaddr_in *sockaddr, char *addr, int port )
{
	struct hostent *hostent;

	/* Attempt name resolution */
	hostent = gethostbyname(addr);
	if ( hostent == NULL )
	{
		printf ( "could not resolve '%s'\n", addr );
		return -1;
	}

	sockaddr->sin_family = hostent->h_addrtype;
	memcpy( (char *) &sockaddr->sin_addr.s_addr, hostent->h_addr_list[0], hostent->h_length );
	sockaddr->sin_port = htons(port);

	/* check that the addr is in mcast range */
	if ( ! IN_MULTICAST(ntohl(sockaddr->sin_addr.s_addr)) )
	{
		printf ( "'%s' not a valid mcast group addr\n", addr );
		return -2;
	}

	return 0;
}

struct client_data *find_client( int naddr, struct client_data *clients, int max )
{
	int i;
	for ( i = 0; i < max ; i++ )
	{
		/*
		printf ("passed %d client %d %d\n", naddr, i, clients[i].naddr );
		*/
		if ( clients[i].naddr == naddr )
		{
			return &clients[i];
		}
	}
	return NULL;
}

int get_options( int argc, char * const argv[], struct options *opts )
{
	int oc;

	while ( 1 )
	{
		oc=getopt( argc, argv, "t:b:m:p:s:c:w:n:h" );
		if ( oc < 0 ) { break; }
		/*
		printf ( "option %c with arg %s\n", oc, optarg );
		*/
		switch (oc)
		{
			case 'h':
				usage( argv[0] );
				break;
			case 's':
				opts->pkt_size = atoi( optarg );
				break;
			case 'c':
				opts->pkts_per_sec = atoi( optarg );
				break;
			case 't':
				opts->ttl = atoi( optarg );
				break;
			case 'b':
				opts->burst = atoi( optarg );
				break;
			case 'w':
				opts->buffer_size = atoi( optarg );
				break;
			case 'n':
				opts->max_clients = atoi( optarg );
				break;
			case 'p':
				opts->port = atoi( optarg );
				break;
			case 'm':
				strcpy( opts->mcast_addr, optarg );
				break;
			case ':':
				printf ("no arg for option %c\n", oc );
				break;
			default:
				printf ("unknown option: %c\n", oc );
				break;
		}

	}
}

#endif
